# -*- coding: utf-8 -*-
from pyramid.threadlocal import get_current_registry, get_current_request
from zope.interface import implements
from penelope.core.models.interfaces import ITicketStore
from penelope.core.models import DBSession
from penelope.trac.api import TracXmlProxy


class TicketStore(object):
    implements(ITicketStore)

    def get_ticket(self, request, project_id, ticket_id):
        from penelope.core.models.dashboard import Project
        project = DBSession().query(Project).get(project_id)
        if request and project:
            for trac in project.tracs:
                proxy = TracXmlProxy('/trac/%s' % trac.trac_name, request=request)
                return proxy.ticket.get(ticket_id)

    def get_tickets_for_project(self, project, request, query=None, limit=None, not_invoiced=False):
        tickets = []
        for trac in project.tracs:
            proxy = TracXmlProxy('/trac/%s' % trac.trac_name, request=request)
            if not query:
                query = [
                    # 'modified=1weekago',
                ]
            if limit:
                query.append('max=%s' % limit)
            else:
                query.append('max=0')
            tickets.extend(proxy.ticket.queryWithDetails('&'.join(query)))

        if not_invoiced:
            cr_ids = [cr.id for cr in project.customer_requests if cr.workflow_state in ['created', 'estimated']]
            tickets = [t for t in tickets if t['cr'] in cr_ids]

        return tickets

    def get_tickets_for_request(self, customer_request, limit=None, request=None):
        cr = customer_request
        return self.get_tickets_for_project(project=cr.project,
                                            query=['customerrequest=%s' % (cr.id)],
                                            limit=limit,
                                            request=request)

    def get_all_ticket_crs(self, project, request):
        """
        returns a mapping of {ticket.id: cr_id}
        """
        import xmlrpclib
        ret = {}

        for trac in project.tracs:
            proxy = TracXmlProxy('/trac/%s' % trac.trac_name, request=request)
            try:
                for ticket_id, cr_id in proxy.ticket.queryAllCustomerRequests():
                    ret[ticket_id] = cr_id
            except xmlrpclib.ProtocolError:
                pass

        return ret

    def get_requests_from_tickets(self, project, ticket_ids, request=None):
        ticket_cr = []
        for trac in project.tracs:
            proxy = TracXmlProxy('/trac/%s' % trac.trac_name, request=request)
            ticket_cr.extend(proxy.ticket.queryCustomerRequestsByTicktes(ticket_ids))
        return ticket_cr

    def add_tickets(self, project, customerrequest, tickets, reporter, notify=False):
        from trac.env import Environment
        from trac.ticket.notification import TicketNotifyEmail
        from trac.ticket.model import Ticket
        from trac.util.text import exception_to_unicode
        from penelope.core.models.dashboard import User

        settings = get_current_registry().settings
        tracenvs = settings.get('penelope.trac.envs')
        request = get_current_request()

        for trac in project.tracs:
            for t in tickets:
                owner = DBSession.query(User).get(t['owner'])
                ticket = {'summary': t['summary'],
                        'description': t['description'],
                        'customerrequest': customerrequest.id,
                        'reporter': reporter.email,
                        'type': 'task',
                        'priority': 'major',
                        'milestone': 'Backlog',
                        'owner': owner.email,
                        'status': 'new'}
                tracenv = Environment('%s/%s' % (tracenvs, trac.trac_name))
                tracenv.abs_href.base = trac.api_uri
                t = Ticket(tracenv)
                t.populate(ticket)
                t.insert()
                if notify:
                    try:
                        tn = TicketNotifyEmail(tracenv)
                        tn.notify(t, newticket=True)
                    except Exception, e:
                        request.add_message('Failure sending notification on creation '
                        'of a ticket #%s: %s' % (t.id, exception_to_unicode(e)), 'error')

ticket_store = TicketStore()



# -*- coding: utf-8 -*-

import colander


def validate_period(node, value):
    ### if a date is required...
    #if not value['date_from'] and not value['date_to']:
    #  raise colander.Invalid(node, 'devi selezionare almeno una data')
    if value['date_from'] and value['date_to'] and value['date_from'] > value['date_to']:
        raise colander.Invalid(node, 'data inizio deve essere <= data fine')



# -*- coding: utf-8 -*-

import xmlrpclib

from copy import deepcopy
from zope.interface import implements
from sqlalchemy import Column
from sqlalchemy import Date
from sqlalchemy import DateTime
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import Interval
from sqlalchemy import Unicode
from sqlalchemy import String
from sqlalchemy import event
from sqlalchemy.orm import relationship, backref

from penelope.core.models.dublincore import dublincore_insert, dublincore_update, DublinCore
from penelope.core.models import Base, DBSession, CustomerRequest
from penelope.core.models import workflow, classproperty
from penelope.core.models.tickets import ticket_store
from penelope.core.models.interfaces import ITimeEntry, IProjectRelated
from penelope.core.security.acl import CRUD_ACL
from penelope.core.lib.helpers import timedelta_as_human_str, timedelta_as_work_days


class TimeEntryException(Exception):
    """Something wrong happened trying to add a time entry to a project"""


class TimeEntry(DublinCore, workflow.Workflow, Base):
    implements(ITimeEntry, IProjectRelated)
    __tablename__ = 'time_entries'

    id = Column(Integer, primary_key=True)
    start = Column(DateTime)
    end = Column(DateTime)
    date = Column(Date, index=True)
    hours = Column(Interval)
    description = Column(Unicode)
    location = Column(Unicode, nullable=False, default=u'RedTurtle')
    ticket = Column(Integer, index=True)
    tickettype = Column(String(25))
    invoice_number = Column(String(10))

    contract_id = Column(String, ForeignKey('contracts.id'), index=True, nullable=True)
    contract = relationship('Contract', uselist=False, backref=backref('time_entries', order_by=id))

    project_id = Column(String, ForeignKey('projects.id'), index=True, nullable=False)
    project = relationship('Project', uselist=False, backref=backref('time_entries', order_by=id))

    @classproperty
    def __acl__(cls):
        acl = deepcopy(CRUD_ACL)
        #add
        acl.allow('role:local_developer', 'new')
        acl.allow('role:local_project_manager', 'new')
        acl.allow('role:external_developer', 'new')
        acl.allow('role:internal_developer', 'new')
        acl.allow('role:secretary', 'new')
        acl.allow('role:project_manager', 'new')
        #view
        acl.allow('role:owner', 'view')
        acl.allow('role:project_manager', 'view')
        acl.allow('role:internal_developer', 'view')
        #edit
        acl.allow('role:project_manager', 'edit')
        acl.allow('role:secretary', 'manage')
        #delete
        isnew = bool(cls.workflow_state == u'new')
        if isnew: #only in new state
            acl.allow('role:owner', 'edit')
            acl.allow('role:owner', 'delete')
        acl.allow('role:project_manager', 'delete')
        #workflow
        acl.allow('role:secretary', 'workflow')
        acl.allow('role:project_manager', 'workflow')
        #listing
        acl.allow('role:project_manager', 'listing')
        return acl

    def __init__(self, *args, **kwargs):
        super(TimeEntry, self).__init__(*args, **kwargs)

    def __str__(self):
        return self.__unicode__().encode('utf8')

    def __unicode__(self):
        return self.plaintext_description

    @property
    def plaintext_description(self):
        return self.description or ''

    @property
    def hours_str(self):
        return timedelta_as_human_str(self.hours)

    @property
    def hours_as_work_days(self):
        return timedelta_as_work_days(self.hours)

    def get_ticket(self, request=None):
        request = request or getattr(self, 'request', None)
        return ticket_store.get_ticket(request, self.project_id, self.ticket)

    def get_ticket_summary(self, request=None):
        ticket = self.get_ticket(request)
        if ticket:
            return ticket[3]['summary']


def new_te_created(mapper, connection, target):
    try:
        trac_ticket = target.get_ticket()
    except xmlrpclib.Error:
        pass
    else:
        if trac_ticket:
            target.tickettype = trac_ticket[3]['type']
            contract_id = DBSession().query(CustomerRequest)\
                                     .get(trac_ticket[3]['customerrequest'])\
                                     .contract_id
            target.contract_id = contract_id


event.listen(TimeEntry, "before_insert", new_te_created)
event.listen(TimeEntry, "before_insert", dublincore_insert)
event.listen(TimeEntry, "before_update", dublincore_update)



import colander
import deform

from deform import ValidationFailure
from deform_bootstrap.widget import ChosenSingleWidget
from pyramid.renderers import get_renderer
from pyramid import httpexceptions as exc

from penelope.core.lib.widgets import SubmitButton, ResetButton, WizardForm
from penelope.core.fanstatic_resources import fastticketing as fastticketing_fanstatic
from penelope.core.models.tickets import ticket_store


class Tickets(colander.SequenceSchema):
    class Ticket(colander.Schema):
          """
              summary: entered by user
              description: entered by user in wiki syntax
              customerrequest: the context,
              reporter: user's email,
              type: 'task',
              priority: 'major',
              milestone: 'Backlog' (choosen for the available ones in trac?)
              owner: user's email
          """
          summary = colander.SchemaNode(typ=colander.String(),
                            widget=deform.widget.TextInputWidget(placeholder=u"Enter the ticket's title"),
                            missing=colander.required,
                            title=u'Summary')
          description = colander.SchemaNode(
                            colander.String(),
                            widget=deform.widget.TextAreaWidget(
                                  cols=60,
                                  rows=5),
                            missing=colander.required,
                            title=u'Description',
                            description=u'use wiki syntax')
          owner = colander.SchemaNode(typ=colander.String(),
                            widget=ChosenSingleWidget(placeholder= u'Select the pal to assign to'),
                            missing=colander.required,
                            title=u'Owner')

    ticket = Ticket(title='')


class FastTicketingSchema(colander.Schema):
    tickets = Tickets()


class FastTicketing(object):
    def __init__(self, context, request):
        self.context = context
        self.request = request

    def render(self):
        result = {}
        result['main_template'] = get_renderer(
               'penelope.core:skins/main_template.pt').implementation()
        result['main'] = get_renderer(
               'penelope.core.forms:templates/master.pt').implementation()

        schema = FastTicketingSchema().clone()
        fastticketing_fanstatic.need()
        form = WizardForm(schema,
                          formid='fastticketing',
                          method='POST',
                          buttons=[
                                 SubmitButton(title=u'Submit'),
                                 ResetButton(title=u'Reset'),
                          ])
        form.bootstrap_form_style = ''
        form['tickets'].widget = deform.widget.SequenceWidget(min_len=1)

        users = set()
        project = self.context.get_instance().project
        for g in getattr(project, 'groups', []):
          for u in g.users:
            users.add(u)
        users.add(project.manager)
        form['tickets']['ticket']['owner'].widget.values = [('', '')] + \
                                      [(str(u.id), u.fullname) for u in list(users)]

        controls = self.request.POST.items()
        if controls != []:
          try:
              appstruct = form.validate(controls)
              self.handle_save(appstruct)
          except ValidationFailure as e:
              result['form'] = e.render()
              return result

        result['form'] = form.render()
        return result

    def handle_save(self, appstruct):
      customerrequest = self.context.get_instance()
      user = self.request.authenticated_user

      ticket_store.add_tickets(project = self.request.model_instance.project, 
                               customerrequest = customerrequest,
                               tickets = appstruct['tickets'],
                               reporter = user,
                               notify=True)

      raise exc.HTTPFound(location=self.request.fa_url('CustomerRequest',
                                                         customerrequest.id))



from zope.interface import Interface

class IApplicationView(Interface):
    pass


class IReportView(Interface):
    "Marker interface for all report views"


class IManageView(Interface):
    "Marker interface for all manage views"


class IBreadcrumbs(Interface):
    "Render my breadcrumbs"


class ISidebar(Interface):
    "Render my sidebar"


class IPorRequest(Interface):
    "Our request marker"


class IRoleFinder(Interface):
    "Adapter that returns roles for context"



from velruse.providers import google
from velruse.exceptions import ThirdPartyFailure
from velruse.exceptions import AuthenticationDenied
from pyramid.httpexceptions import HTTPFound


def includeme(config):
    settings = config.registry.settings
    store, realm = setup_openid(config)
    consumer = RedturtleConsumer(
        storage=store,
        realm=realm,
        process_url='redturtle_process',
        oauth_key=settings.get('velruse.google.consumer_key'),
        oauth_secret=settings.get('velruse.google.consumer_secret'),
        request_attributes=settings.get('request_attributes')
    )
    config.add_route("redturtle_login", "/redturtle/login")
    config.add_route("redturtle_process", "/redturtle/process",
                     use_global_views=True,
                     factory=consumer.process)
    config.add_view(consumer.login, route_name="redturtle_login", request_method="POST")


def setup_openid(config):
    settings = config.registry.settings
    store = getattr(config.registry,'velruse.openid_store', None)
    if not store and 'velruse.openid.store' not in settings:
        raise Exception("Missing 'velruse.openid.store' in config settings.")
    if not store:
        store = config.maybe_dotted(settings['velruse.openid.store'])()
        setattr(config.registry,'velruse.openid_store',store)
    realm = settings['velruse.openid.realm']
    return store, realm


class RedturtleConsumer(google.GoogleConsumer):

    def _lookup_identifier(self, request, identifier):
        """Return the Google OpenID directed endpoint"""
        domain = request.registry.settings.get('penelope.core.google_domain')
        return "https://www.google.com/accounts/o8/site-xrds?hd=%s" % domain

    def login(self, context, request):
        return super(RedturtleConsumer, self).login(request)

    def process(self, request):
        try:
            return super(RedturtleConsumer, self).process(request)
        except ThirdPartyFailure:
            raise HTTPFound(location='/')
        except AuthenticationDenied:
            request.add_message('Authentication from google has failed. Try again.', 'error')
            raise HTTPFound(location='/')



# -*- coding: utf-8 -*-

import datetime
import re

from pyramid.view import view_config
from pyramid.response import Response

from penelope.core.api import timeentry_crstate_validation_errors
from penelope.core.lib.helpers import unicodelower
from penelope.core.lib.helpers import time_chunks
from penelope.core.models import DBSession
from penelope.core.models.dashboard import Project
from penelope.core.models.tp import TimeEntry
from penelope.core.models.tickets import ticket_store


class SmartAdd(object):
    def __init__(self, context, request):
        self.context = context
        self.request = request


    @view_config(name='smartadd_projects', renderer='json')
    def smartadd_projects(self):
        """
        Returns a json list of project tags
        """
        qry = self.request.filter_viewables(DBSession.query(Project).filter(Project.active))
        projects = sorted((u'%s - %s' % (p.name, p.customer.name) for p in qry), key=unicodelower)
        return projects


    @view_config(name='smartadd_tickets', renderer='json')
    def smartadd_tickets(self):
        """
        Returns a json list of ticket tags
        """
        qry = self.request.filter_viewables(DBSession.query(Project).filter(Project.active))
        projects = dict(('%s - %s' % (p.name, p.customer.name), p.id) for p in qry)
        text = self.request.params['text']

        parser = SmartAddParser(text,
                                projects=projects,
                                request=self.request)

        if parser.TAG_TICK in parser.unparsed:
            ticket_search = parser.unparsed.split(parser.TAG_TICK)[-1]
        else:
            ticket_search = ''


        if ticket_search.isdigit():
            # ^= is startswith (http://trac.edgewall.org/wiki/TracQuery)
            query = ['id^=%s' % ticket_search]
        elif ticket_search:
            query = ['summary~=%s' % ticket_search]
        else:
            query = None

        if parser.project_id:
            project = DBSession.query(Project).get(parser.project_id)
            tickets = ticket_store.get_tickets_for_project(
                            project=project,
                            request=self.request,
                            query=query,
                            not_invoiced=True,
                            limit=15)
        else:
            tickets = []

        return [
            "%(id)s %(summary)s" % ticket
            for ticket in tickets
            if ticket['resolution'] not in ('invalid', 'duplicate')
            ]


    @view_config(name='smartadd_submit', request_method='POST')
    def smartadd_submit(self):
        """
        Receives a line of smart-add and performs validation/insertion.
        """
        projects = dict(
                ('%s - %s' % (p.name, p.customer.name), p.id)
                for p in self.request.filter_viewables(DBSession.query(Project).filter(Project.active))
                )

        def ticket_provider(project_id):
            if project_id:
                project = DBSession.query(Project).get(project_id)
                return [
                        t['id']
                        for t in ticket_store.get_tickets_for_project(project=project,
                                                                      not_invoiced=True,
                                                                      request=self.request)
                        ]

        parser = SmartAddParser(unicode(self.request.body, 'utf8', 'ignore'),
                                projects=projects,
                                available_tickets=ticket_provider,
                                request=self.request)

        errors = parser.validation_errors()
        if errors:
            # XXX register appropriate exception handler
            return Response(' - '.join(errors), status_int=400)

        pte = parser.parsed_time_entry
        parsed_tickets = pte['tickets']
        ticket_summaries = []

        entry_durations = list(time_chunks(pte['hours'], len(parsed_tickets)))

        for parsed_ticket, duration in zip(parsed_tickets, entry_durations):
            date = pte.get('date') or datetime.date.today()

            te = TimeEntry(date = date,
                           start = pte['start'],
                           end = pte['start'],
                           description = pte['description'],
                           ticket = parsed_ticket,
                           project_id = pte['project_id'],
                           hours = duration,
                           )
            te.request = self.request #bind for user calculation
            DBSession.add(te)
            # retrieve ticket descriptions (another trip to the store..)
            ticket_summaries.append(
                    '#%s (%s)' % (te.ticket, ticket_store.get_ticket(self.request, te.project_id, te.ticket)[3]['summary'])
                )

        return Response(u'Added to ticket(s) %s' % ', '.join(ticket_summaries))


class SmartAddParser(object):
    TAG_PROJ = '@'
    TAG_TIME = '^'
    TAG_TICK = '#'
    TAG_DATE = '!'

    project_id = None
    ticket = None
    hours = None
    unparsed = None

    def __init__(self, text, projects=None, available_tickets=None, request=None):
        self.project_id, text = self.parse_project(text, projects or {})
        self.tickets, text = self.parse_tickets(text, self.project_id, available_tickets or {})
        self.hours, text = self.parse_hours(text)
        self.date, text = self.parse_date(text)
        self.unparsed = text.strip()
        self.request = request


    @classmethod
    def parse_project(cls, text, projects):
        found_match = None
        found_string = ''
        found_start = 1e6

        # case insensitive search and lookup
        projects = dict((p[0].lower(), p[1]) for p in projects.iteritems())

        for project_name, project_id in projects.items():
            m = re.search('%s(%s)' % (re.escape(cls.TAG_PROJ), re.escape(project_name)), text, flags=re.IGNORECASE)
            if m and ((found_start > m.start()) or len(project_name) > len(found_string)):
                found_match = m
                found_string = m.groups()[0].lower()
                found_start = m.start()

        if found_string:
            text = cls._removematch(text, found_match)

        return projects.get(found_string), text


    @classmethod
    def _removematch(self, text, m):
        """
        Returns the text string minus the match.
        """
        start, end = m.span()
        head, tail = text[:start], text[end:]
        if tail.startswith(' '):
            tail = tail[1:]
        return head + tail


    @classmethod
    def _removematchingticket(self, text, ticket_id):
        """
        Returns the text string minus the matching ticket occurrences.
        """
        while True:
            match = re.search('%s(%s)(?=\D|$)' % (re.escape(self.TAG_TICK), ticket_id), text)
            if not match:
                break
            text = self._removematch(text, match)

        return text


    def _matching_tickets(self, text):
        """
        Returns the matching ticket numbers still inside 'text'.
        """
        return re.findall('%s(\d+)(?=\D|$)' % re.escape(self.TAG_TICK), text)


    def parse_tickets(self, text, project_id, available_tickets):

        if callable(available_tickets):
            available_tickets = available_tickets(project_id)

        if available_tickets is None:
            available_tickets = []

        available_tickets = map(str, available_tickets)

        found_tickets = []
        for ticket_id in self._matching_tickets(text):
            if ticket_id in available_tickets and ticket_id not in found_tickets:
                found_tickets.append(ticket_id)

        for ticket_id in found_tickets:
            text = self._removematchingticket(text, ticket_id)

        return found_tickets, text


    def parse_hours(self, text):
        m = re.search('%s(?P<hh>\d\d*):(?P<mm>\d\d)(?=\D|$)' % re.escape(self.TAG_TIME), text)
        if m:
            gd = m.groupdict()
            text = self._removematch(text, m)
            return datetime.timedelta(hours=int(gd['hh']), minutes=int(gd['mm'])), text

        m = re.search('%s(?P<hh>\d\d*)(?=[^:\d]|$)' % re.escape(self.TAG_TIME), text)
        if m:
            gd = m.groupdict()
            text = self._removematch(text, m)
            return datetime.timedelta(hours=int(gd['hh'])), text

        m = re.search('%s:(?P<mm>\d\d*)(?=\D|$)' % re.escape(self.TAG_TIME), text)
        if m:
            gd = m.groupdict()
            text = self._removematch(text, m)
            return datetime.timedelta(minutes=int(gd['mm'])), text

        return None, text


    def parse_date(self, text):
        m = re.search('%s(today|yesterday)' % re.escape(self.TAG_DATE), text)
        if m:
            day = m.groups()[0]
            text = self._removematch(text, m)
            if day == 'today':
                return datetime.date.today(), text
            elif day == 'yesterday':
                return (datetime.date.today() - datetime.timedelta(days=1)), text

        return None, text


    def validation_errors(self):
        errors = []

        if self.project_id is None:
            if self.TAG_PROJ in self.unparsed:
                errors.append(u'Project not found')
            else:
                errors.append(u'Project is missing')

        if self.hours is None:
            if self.TAG_TIME in self.unparsed:
                errors.append(u'Could not parse time')
            else:
                errors.append(u'Time is missing')

        if self.date is None and self.TAG_DATE in self.unparsed:
            errors.append(u'Cannot parse date')

        missing_tickets = self._matching_tickets(self.unparsed)
        if missing_tickets:
            errors.append(u'{ticket_label} {ticket_numbers} not found or customer request already invoiced'.format(
                                        ticket_label = ['Ticket', 'Tickets'][len(missing_tickets)>1],
                                        ticket_numbers = ', '.join('#'+ticket_id for ticket_id in missing_tickets)),
                                        )
        elif not self.tickets:
            errors.append(u'Ticket is missing')

        if not self.unparsed:
            errors.append(u'Description is empty')

        if self.request:
            errors.extend(timeentry_crstate_validation_errors(self.project_id, self.tickets, request=self.request))

        return errors


    @property
    def parsed_time_entry(self):
        return {
                'start': None,
                'end': None,
                'hours': self.hours,
                'description': self.unparsed,
                'location': None,
                'tickets': self.tickets or None,
                'project_id': self.project_id,
                'date': self.date,
                }



# -*- coding: utf-8 -*-

import colander
import datetime
import functools
import itertools
import logging
import operator
import sqlalchemy as sa
import webhelpers

from deform import ValidationFailure
from pyramid.renderers import render
from pyramid.view import view_config

from penelope.core.lib.helpers import ticket_url
from penelope.core.lib.widgets import SearchButton, PorInlineForm
from penelope.core.models import DBSession, TimeEntry
from penelope.core.models.tp import timedelta_as_human_str, timedelta_as_work_days
from penelope.core.reports import fields
from penelope.core.reports.favourites import render_saved_query_form
from penelope.core.reports.queries import qry_active_projects
from penelope.core.reports.validators import validate_period

log = logging.getLogger(__name__)


def yesterday():
    return datetime.date.today() - datetime.timedelta(1)


@colander.deferred
def deferred_today(node, kw):
    default_date = kw.get('today')
    if default_date is None:
        default_date = datetime.date.today()
    return default_date


@colander.deferred
def deferred_yesterday(node, kw):
    default_date = kw.get('yesterday')
    if default_date is None:
        default_date = yesterday()
    return default_date


class MyEntriesReport(object):
    def __init__(self, context, request):
        self.context = context
        self.request = request


    class MyEntriesSchema(colander.MappingSchema):
        project_id = fields.project_id.clone()
        date_from = fields.date_from.clone()
        date_from.default = deferred_yesterday
        date_from.missing = deferred_yesterday
        date_to = fields.date_to.clone()
        date_to.default = deferred_today
        date_to.missing = deferred_today
        searchtext = fields.searchtext.clone()


    def search(self, limit, author_id, project_id, date_from, date_to, searchtext):
        qry = DBSession.query(TimeEntry)

        qry = qry.filter(TimeEntry.author_id==author_id)

        if project_id is not colander.null:
            qry = qry.filter(TimeEntry.project_id==project_id)

        if date_from is not colander.null:
            qry = qry.filter(TimeEntry.date>=date_from)

        if date_to is not colander.null:
            qry = qry.filter(TimeEntry.date<=date_to)

        if searchtext is not colander.null:
            qry = qry.filter(TimeEntry.description.ilike(u'%{0}%'.format(searchtext)))

        qry = qry.order_by(sa.desc(TimeEntry.date), sa.desc(TimeEntry.start), sa.desc(TimeEntry.creation_date))

        if limit:
            qry = qry.limit(limit)

        qry = self.request.filter_viewables(qry)
        entries_by_date = []
        entries_count = 0
        
        for k, g in itertools.groupby(qry, operator.attrgetter('date')):
            g = list(g)
            entries_by_date.append((k, g))
            entries_count += len(g)

        return entries_by_date, entries_count


    @view_config(name='report_my_entries', route_name='reports', renderer='skin', permission='reports_my_entries')
    def __call__(self):
        schema = self.MyEntriesSchema(validator=validate_period).clone()\
                     .bind(yesterday=yesterday(),
                           today=datetime.date.today())
        limit = 1000
        projects = qry_active_projects().all()
        form = PorInlineForm(schema,
                             formid='my_entries',
                             method='GET',
                             buttons=[
                                 SearchButton(title=u'Search'),
                             ])

        form['project_id'].widget.values = [('', '')] + sorted([(str(p.id), ' / '.join([p.customer.name, p.name])) for p in projects],
                                                               key=lambda x: x[1].lower())

        result_table = ''
        controls = self.request.GET.items()

        try:
            appstruct = form.validate(controls)
        except ValidationFailure as e:
            return {
                    'form': e.render(),
                    'saved_query_form': render_saved_query_form(self.request),
                    'result_table': None
                    }
        
        current_uid = self.request.authenticated_user.id
        entries_by_date, entries_count = self.search(author_id=current_uid,
                                                     limit=limit,
                                                     **appstruct)

        highlight = functools.partial(webhelpers.html.tools.highlight,
                                      phrase=appstruct['searchtext'])

        delta0 = datetime.timedelta()
        delta_tot = sum([sum((e.hours for e in x[1]), delta0) for x in entries_by_date], delta0)

        human_tot = timedelta_as_human_str(delta_tot)
        days_tot = timedelta_as_work_days(delta_tot)

        result_table = render('penelope.core:reports/templates/my_entries_results.pt',
                              {
                                  'entries_by_date': entries_by_date,
                                  'entries_count': entries_count,
                                  'highlight': highlight,
                                  'human_tot': human_tot,
                                  'days_tot': days_tot,
                                  'datetime': datetime,
                                  'ticket_url': ticket_url,
                                  'timedelta_as_human_str': timedelta_as_human_str,
                              },
                              request=self.request)

        return {
                'form': form.render(appstruct=appstruct),
                'saved_query_form': render_saved_query_form(self.request),
                'result_table': result_table,
                }



# -*- coding: utf-8 -*-

from datetime import timedelta
from simplejson import dumps
import xmlrpclib

from webhelpers.html import HTML, literal
from formalchemy import helpers as h
from formalchemy import fatypes, validators
from formalchemy import FieldSet, Grid;Grid #pyflakes
from formalchemy import fields

from fa.jquery.fanstatic_resources import fa_js
from fa.jquery.renderers import DateFieldRenderer, default_renderers

from penelope.core.lib.helpers import listwrap, ticket_url, unicodelower
from penelope.core.models import DBSession, User, CustomerRequest
from penelope.core.models.tickets import ticket_store


def _query_options(L):
    if hasattr(L, '_mapper_adapter_map') and \
                                         User in L._mapper_adapter_map.keys():
        L = L.filter_by(active=True)
    return [(fields._stringify(item), fields._pk(item)) for item in L]
fields._query_options = _query_options


def grooming_label_renderer(renderer=fields.SelectFieldRenderer):
    class Renderer(renderer):
        def render_readonly(self, options=None, **kwargs):
            if self.raw_value == CustomerRequest.PLACEMENT_BACKLOG:
                return h.literal(u'<span class="label label-important">Backlog</span>')
            elif self.raw_value == CustomerRequest.PLACEMENT_GROOMING:
                return h.literal(u'<span class="label label-warning">Grooming</span>')
            elif self.raw_value == CustomerRequest.PLACEMENT_BOARD:
                return h.literal(u'<span class="label label-success">Board</span>')
            else:
                return h.literal(self.raw_value)

        def render(self, **kwargs):
            return h.select(self.name, self.value, [('Backlog', '0'), ('Grooming', '1'), ('Board', '2')], **kwargs)

    return Renderer

class UrlRenderer(fields.TextFieldRenderer):
    def render_readonly(self, **kwargs):
        return HTML.A(self.value, href=self.value, target="_blank")


class TicketRenderer(fields.IntegerFieldRenderer):
    def render_readonly(self, **kwargs):
        request = self.request
        te = request.model_instance

        if not request.has_permission('view', te.project):
            return '#%s' % te.ticket

        try:
            ticket_data = ticket_store.get_ticket(request, te.project_id, te.ticket)
        except (xmlrpclib.Fault, xmlrpclib.Error):
            ticket_data = None

        if ticket_data:
            ticket_id, dummy_time, dummy_changetime, tkt = ticket_data
            ret = HTML.A(u'#%s - %s' % (ticket_id, tkt['summary']),
                         href=ticket_url(request, te.project, te.ticket))
            cr_id = tkt['customerrequest']
            if cr_id:
                cr = DBSession.query(CustomerRequest).get(cr_id)
                ret += HTML.BR() + HTML.SPAN('CR: ') + HTML.A(cr.name, href='/admin/CustomerRequest/%s' % cr_id)
            return ret
        else:
            return HTML.SPAN(u'#%s - ' % te.ticket, HTML.SPAN(u'NOT FOUND', class_='label label-important'))


class RelationRenderer(fields.SelectFieldRenderer):
    """Make relation linkable in readonly mode
       and aware of js.chosen()"""

    def render(self, options, **kwargs):
        if callable(options):
            L = fields._normalized_options(options(self.field.parent))
            if not self.field.is_required() and not self.field.is_collection:
                L.insert(0, self.field._null_option)
        else:
            L = list(options)
        if len(L) > 0:
            if len(L[0]) == 2:
                L = sorted([(k, self.stringify_value(v)) for k, v in L], key=lambda x:unicodelower(x[0]))
            else:
                L = sorted([fields._stringify(k) for k in L], key=unicodelower)
        return h.select(self.name, self.value, L, class_='i-can-haz-chzn-select', **kwargs) + \
               h.literal("<script>$('#%s').chosen()</script>" % self.name)

    def render_readonly(self, options=None, **kwargs):
        value = self.raw_value
        if value is None:
            return ''

        if not isinstance(value, list):
            if self.request.has_permission('view', value):
                return h.content_tag('a',
                    self.stringify_value(value, as_html=True),
                    href=self.request.fa_url(value.__class__.__name__, fields._pk(value)))
            else:
                return h.content_tag('span', self.stringify_value(value, as_html=True))
        else:
            html = []
            for item in value:
                if self.request.has_permission('view', item):
                    html.append(h.content_tag('a',
                                self.stringify_value(item, as_html=True),
                                href=self.request.fa_url(item.__class__.__name__, fields._pk(item))))
                else:
                    html.append(h.content_tag('span', self.stringify_value(item, as_html=True)))
            return h.literal(',&nbsp;').join(html)


class PORDateFieldRenderer(DateFieldRenderer):
    """Make sure we are using fa_jquery resources"""

    def render(self, **kwargs):
        fa_js.need()
        return super(PORDateFieldRenderer, self).render(**kwargs)


class IntervalFieldRenderer(fields.IntervalFieldRenderer):
    """Use http://jqueryui.com/demos/datepicker/"""

    jq_options = dict(minuteGrid=10, hourGrid=4, showButtonPanel=False, timeFormat='hh:mm')
    template = """<input type="text" autocomplete="off" size="10" value="%(value)s" id="%(name)s" name="%(name)s" />
    <script type="text/javascript">$('#%(name)s').timepicker(%(jq_options)s);</script>"""

    def stringify_timedelta(self, value):
        if not isinstance(value, timedelta):
            return value

        seconds = value.seconds
        days = value.days
        hours, remainder = divmod(seconds, 3600)
        hours += (days * 24)
        minutes, seconds = divmod(remainder, 60)
        return "%d:%02d" % (hours, minutes)

    def stringify_value(self, v, as_html=False):
        if as_html or not v:
            return super(IntervalFieldRenderer, self).stringify_value(v, as_html=as_html)
        return self.stringify_timedelta(v)

    def render_readonly(self, **kwargs):
        value = self.raw_value
        return self.stringify_timedelta(value)

    def render(self, **kwargs):
        value = self.raw_value or ''
        try:
            value = self.stringify_timedelta(value)
        except validators.ValidationError:
            value = '00:00'
        options = self.jq_options.copy()
        kwargs.update(
            name=self.name,
            value=value,
            jq_options=dumps(options),
        )
        return literal(self.template % kwargs)

    def _deserialize(self, data):
        try:
            hours, minutes = [int(s) for s in data.split(':')]
        except ValueError:
            raise validators.ValidationError('Not a proper interval')
        return timedelta(hours=hours, minutes=minutes)


FieldSet.default_renderers.update(default_renderers)
FieldSet.default_renderers['dropdown'] = RelationRenderer
FieldSet.default_renderers[fatypes.Date] = PORDateFieldRenderer
FieldSet.default_renderers[fatypes.Interval] = IntervalFieldRenderer


class ProjectRelationRenderer(RelationRenderer):
    """
    Renders project names without the customer
    """

    def _stringify_project(self, project, as_html):
        if as_html:
            return h.content_tag('a',
                                 project.name,
                                 href=self.request.fa_url(project.__class__.__name__, fields._pk(project)))
        else:
            return project.name

    def stringify_value(self, value, as_html=False):
        return literal(',&nbsp;').join(
                        self._stringify_project(project, as_html=as_html)
                        for project in listwrap(value)
                    )




import zope.component
from zope.interface import Interface
from fa.bootstrap import actions

from penelope.core.interfaces import IBreadcrumbs, IReportView
from penelope.core.models.interfaces import IPorModel, IProjectRelated, IKanbanBoard


gsm = zope.component.getGlobalSiteManager()


class PORBreadcrumbAction(actions.BreadcrumbAction):
    body = u'''<li tal:attributes="class action.isActive(request) and \'active\' or \'\'">
                    <a tal:attributes="%(attributes)s">${content}</a>
                    <span tal:condition="not action.isActive(request)" class="divider">/</span>
               </li>'''


class BreadcrumbRenderer(object):
    def __init__(self, context):
        self.context = context
        self.breadcrumbs = actions.actions.Actions()
        self.breadcrumbs.append(PORBreadcrumbAction('home',
            content='Home',
            attrs=dict(href="request.application_url")))

    @property
    def current_id(self):
        if self.context.request:
            return self.context.request.view_name or getattr(self.context.request, 'model_name', None)
        else:
            return ''

    @property
    def current_title(self):
        return self.current_id.replace('_', ' ').title()

    def render(self, request):
        if self.current_id:
            self.breadcrumbs.append(PORBreadcrumbAction(self.current_id,
                content=self.current_title,
                attrs=dict(href="request.path_url")))
        if len(self.breadcrumbs)>1:
            return self.breadcrumbs.render(request)
        else:
            return ''

gsm.registerAdapter(BreadcrumbRenderer, (Interface,), IBreadcrumbs)


class BreadcrumbReportRenderer(BreadcrumbRenderer):
    def render(self, request):
        if self.current_id != 'index':
            self.breadcrumbs.append(PORBreadcrumbAction('reports',
                content='Reports',
                attrs=dict(href="'%s/reports/index' % request.application_url")))
        return super(BreadcrumbReportRenderer, self).render(request)

gsm.registerAdapter(BreadcrumbReportRenderer, (IReportView,), IBreadcrumbs)


class BreadcrumbCrudRenderer(BreadcrumbRenderer):
    @property
    def current_id(self):
        return self.context.id

    @property
    def current_title(self):
        return unicode(self.context)

gsm.registerAdapter(BreadcrumbCrudRenderer, (IPorModel,), IBreadcrumbs)


class BreadcrumbKanbanRenderer(BreadcrumbCrudRenderer):
    def render(self, request):
        self.breadcrumbs.append(PORBreadcrumbAction('Kanbanboard',
            content='Kanbanboard',
            attrs=dict(href="'%s/admin/KanbanBoard' % request.application_url")))
        return super(BreadcrumbKanbanRenderer, self).render(request)

gsm.registerAdapter(BreadcrumbKanbanRenderer, (IKanbanBoard,), IBreadcrumbs)


class BreadcrumbProjectRelatedRenderer(BreadcrumbRenderer):
    def render(self, request):
        project = self.context.project

        if project.customer:
            self.breadcrumbs.append(PORBreadcrumbAction(project.customer.id,
                content=unicode(project.customer),
                attrs=dict(href="request.fa_url('Customer','%s')" % project.customer.id)))

        if project != self.context:
            self.breadcrumbs.append(PORBreadcrumbAction(project.id,
                content=unicode(project.name),
                attrs=dict(href="request.fa_url('Project','%s')" % project.id)))

            if request.model_name in ('CustomerRequest','Application','Group'):
                model_name = self.context.project_related_label
                self.breadcrumbs.append(PORBreadcrumbAction(model_name,
                    content=unicode(model_name),
                    attrs=dict(href="request.fa_url('Project','%s', '%s')" % (project.id, self.context.project_related_id))
                ))

        if project == self.context:
            name = project.name
        else:
            name = unicode(self.context)
        self.breadcrumbs.append(PORBreadcrumbAction(self.context.id,
            content=name,
            attrs=dict(href="request.path_url")))

        return self.breadcrumbs.render(request)

gsm.registerAdapter(BreadcrumbProjectRelatedRenderer, (IProjectRelated,), IBreadcrumbs)





import datetime

from zope.deprecation import deprecated

from sqlalchemy.orm import exc as orm_exc
from pyramid_rpc.jsonrpc import jsonrpc_method

from penelope.core.lib.helpers import time_parse

from penelope.core.models import DBSession
from penelope.core.models.tp import TimeEntry, timedelta_as_human_str
from penelope.core.models.dashboard import User, Customer, Project, CustomerRequest
from penelope.core.models.tickets import ticket_store


def decode_string(entry):
    if not isinstance(entry, basestring):
        entry = entry.decode('utf-8')
    return entry


def timeentry_crstate_validation_errors(project_id, tickets, request):
    # XXX this check is deactivated for now (see #312)
    return []

    project = DBSession.query(Project).get(project_id)

    customer_requests = ticket_store.get_requests_from_tickets(project, tickets, request=request)

    for ticket_id, cr_id in customer_requests:
        cr = DBSession.query(CustomerRequest).get(cr_id)
        if cr.workflow_state != 'estimated':
            return ['Customer Request is not estimated']

    return []


@jsonrpc_method(endpoint='DashboardAPI')
def get_user_by_login(request, login):
    """
    Search for user using his login name
    """
    return get_user_by_email(request, login)


deprecated('get_user_by_login',
           'Use of get_user_by_login is deprecated. Use get_user_by_email instead.')


@jsonrpc_method(endpoint='DashboardAPI')
def get_user_by_email(request, email):
    """
    This method search for user using his email address
    """
    session = DBSession()
    if not isinstance(email, basestring):
        return {
                'status:': False,
                'message': u'Email parameter must be a string!',
                }
    try:
        user = session.query(User).filter_by(email=email).one()
    except orm_exc.NoResultFound:
        return {
                'status': False,
                'message': u'No user found in db for %s mail address' % email,
                }

    return {
            'status': True,
            'message': u'User found.',
            'email': user.email,
            'login': user.login,
            'openids': [x.openid for x in user.openids]
            }


@jsonrpc_method(endpoint='DashboardAPI')
def get_user_by_openid(request, openid):
    """
    This method search for user using one of the possible user openids
    """
    session = DBSession()

    if not isinstance(openid, basestring):
        return {
                'status:': False,
                'message': u'Openid parameter must be a string!',
                }

    try:
        user = session.query(User).join('openids').filter_by(openid=openid).one()
    except orm_exc.NoResultFound:

        return {
                'status': False,
                'message': u'No user found in db for %s openid' % openid,
                }

    return {
            'status': True,
            'message': u'User found.',
            'email': user.email,
            'login': user.login,
            'openids': [x.openid for x in user.openids],
            }


@jsonrpc_method(endpoint='DashboardAPI')
def get_customer_by_name(request, customer_name):
    """
    This method search for customer by name
    """
    session = DBSession()

    if not isinstance(customer_name, basestring):
        return {
                'status:': False,
                'message': u'Customer name parameter must be a string!',
                }

    try:
        customer = session.query(Customer).filter_by(name=customer_name).one()
    except orm_exc.NoResultFound:
        return {
                'status': False,
                'message': u'No customer found in db for %s name' % customer_name,
                }

    return {
            'status': True,
            'message': u'Customer found.',
            'name': customer.name,
            'projects': [x.name for x in customer.projects],
            }


@jsonrpc_method(endpoint='DashboardAPI')
def get_project_by_id(request, project_id):
    """
    This method search for project by name
    """
    session = DBSession()

    if not isinstance(project_id, basestring):
        return {
                'status:': False,
                'message': u'Project id parameter must be a string!',
                }

    try:
        project = session.query(Project).get(project_id)
    except orm_exc.NoResultFound:
        return {
                'status': False,
                'message': u'No project found in db for %s name' % project_id
                }

    return {
            'status': True,
            'message': u'Project found.',
            'name': project.name,
            'id': project.id,
            'customer': project.customer.name,
            'applications': [x.name for x in project.applications],
            'customer_requests': [
                                    {
                                        'id': cr.id,
                                        'name': cr.name,
                                        'active': cr.placement == cr.PLACEMENT_BOARD
                                    }
                                    for cr in project.customer_requests
                                    ]
            }


@jsonrpc_method(endpoint='DashboardAPI')
def get_project_by_name(request, project_name):
    """
    This method search for project by name
    """
    session = DBSession()

    if not isinstance(project_name, basestring):
        return {
                'status:': False,
                'message': u'Project name parameter must be a string!',
                }

    try:
        project = session.query(Project).filter_by(name=project_name).one()
    except orm_exc.NoResultFound:
        return {
                'status': False,
                'message': u'No project found in db for %s name' % project_name,
                }

    return {
            'status': True,
            'message': u'Project found.',
            'name': project.name,
            'id': project.id,
            'customer': project.customer.name,
            'applications': [x.name for x in project.applications],
            'customer_requests': [(x.id, x.name,) for x in project.customer_requests],
            }


@jsonrpc_method(endpoint='DashboardAPI')
def time_entry_total_duration(request, date):
    try:
        date = datetime.datetime.strptime(date, '%Y-%m-%d')
    except ValueError:
        return {
                'status': False,
                'message': 'bad date format',
                }

    qry = DBSession.query(TimeEntry).filter(TimeEntry.date==date)
    qry = qry.filter(TimeEntry.author_id==request.authenticated_user.id)
    duration = sum((te.hours for te in qry), datetime.timedelta(0))

    # we can return past timeentry descriptions here

    return {
            'status': True,
            'duration': timedelta_as_human_str(duration),
            }


@jsonrpc_method(endpoint='DashboardAPI')
def create_new_simple_time_entry(request, entry_ticket, entry_date, entry_hours,
                                 entry_description, entry_location, entry_project):
    """
    Time entry creation: simple time entry case
    """

    try:
        entry_location = decode_string(entry_location)
        entry_description = decode_string(entry_description)
        entry_ticket = decode_string(entry_ticket)
        entry_time_delta = time_parse(entry_hours,
                                      minimum=datetime.timedelta(seconds=1*60),
                                      maximum=datetime.timedelta(seconds=16*60*60))
    except Exception, e:
        return {
                'status': False,
                'message': str(e),
                }

    if not entry_description:
        return {
                'status': False,
                'message': u'Description is required.',
                }

    try:
        entry_date = datetime.datetime.strptime(entry_date, '%Y-%m-%d')
    except ValueError, e:
        return {
                'status': False,
                'message': str(e),
                }


    if not entry_time_delta:
        return {
                'status': False,
                'message': 'Duration is required',
                }


    session = DBSession()
    project = session.query(Project).filter(Project.id==entry_project).first()
    if not project:
        return {
                'state': False,
                'message': u'Not able to get the project with id %s' % entry_project,
                }

    crstate_errors = timeentry_crstate_validation_errors(entry_project, [entry_ticket], request)
    if crstate_errors:
        return {
                'state': False,
                'message': '\n'.join(crstate_errors),
                }

    time_entry = TimeEntry(date = entry_date,
                           hours = entry_time_delta,
                           location = entry_location,
                           description = entry_description,
                           ticket = entry_ticket)
    time_entry.request = request        # bind for user lookup

    time_entry.project_id = entry_project
    session.add(time_entry)
    session.flush()

    return {
            'status': True,
            'message': u'Correctly added time entry %s for %s ticket #%s' % (time_entry.id, entry_project, entry_ticket),
            }


@jsonrpc_method(endpoint='DashboardAPI')
def create_new_advanced_time_entry(request, entry_ticket, entry_start, entry_end,
                                   entry_description, entry_location, entry_project):
    """
    Time entry creation: simple time entry case
    """

    try:
        entry_location = decode_string(entry_location)
        entry_description = decode_string(entry_description)
        entry_ticket = decode_string(entry_ticket)
    except Exception, e:
        return {
                'status': False,
                'message': str(e),
                }

    try:
        entry_start = datetime.datetime.strptime(entry_start, '%Y-%m-%d %H:%M:%S')
    except ValueError, e:
        return {
                'status': False,
                'message': str(e),
                }

    try:
        entry_end = datetime.datetime.strptime(entry_end, '%Y-%m-%d %H:%M:%S')
    except ValueError, e:
        return {
                'status': False,
                'message': str(e),
                }

    entry_date = datetime.date.today()
    session = DBSession()
    project = session.query(Project).filter(Project.id==entry_project).first()
    if not project:
        return {
                'state': False,
                'message': u'Not able to get the project with id %s' % entry_project,
                }

    time_entry = TimeEntry(date = entry_date,
                           start = entry_start,
                           end = entry_end,
                           location = entry_location,
                           description = entry_description,
                           ticket = entry_ticket)

    time_entry.project_id = entry_project
    session.add(time_entry)
    session.flush()

    return {
            'status': True,
            'message': u'Correctly added time entry %s for %s ticket #%s' % (time_entry.id, entry_project, entry_ticket)
            }



# -*- coding: utf-8 -*-
from pyramid.view import view_config

from penelope.core.lib.helpers import unicodelower
from penelope.core.reports.views import ReportContext; ReportContext
from penelope.core.reports.queries import filter_users_with_timeentries, qry_active_projects
from penelope.core.models import DBSession, Project, SavedQuery, User


@view_config(name='index', route_name='reports', renderer='skin', permission='reports_index')
def report_index(context, request):
    users = DBSession.query(User).order_by(User.fullname)
    users = filter_users_with_timeentries(users)
    projects = sorted(request.filter_viewables(qry_active_projects()), key=unicodelower)
    customers = sorted(set(p.customer for p in projects), key=unicodelower)

    current_uid = request.authenticated_user.id
    saved_queries = DBSession.query(SavedQuery).filter(SavedQuery.author_id==current_uid)
    return {
            'users': users,
            'customers': customers,
            'projects': projects,
            'saved_queries': saved_queries.all()
            }



@view_config(name='project_tree', route_name='reports', renderer='json', xhr=True)
def project_tree(context, request):
    """
    This view is used by the customer-project-request javascript filter.
    """
    all_projects = request.filter_viewables(DBSession.query(Project))

    all_projects = [
            project for project in request.filter_viewables(qry_active_projects())
            if request.has_permission('reports_all_entries_for_project', project)
            ]

    customers = request.filter_viewables(set(p.customer for p in all_projects if p.active))

    return [
            {
                'id': str(c.id),
                'name': c.name,
                'projects': [
                    {
                        'id': str(p.id),
                        'name': p.name,
                        'customer_requests': [
                            {
                                'id': str(cr.id),
                                'name': cr.name,
                                }
                            for cr in p.customer_requests
                            ],
                        }
                    for p in c.projects if p in all_projects
                    ]
                }
            for c in customers
            ]




# -*- coding: utf-8 -*-

import datetime
import operator
import re

SECS_IN_HR = 60.0*60.0
WORK_HOURS_IN_DAY = 8.0


def timedelta_as_work_days(td):
    return (td.days*24.0 + td.seconds/SECS_IN_HR) / WORK_HOURS_IN_DAY


def timedelta_as_human_str(td, seconds=False):
    """
    Formats a timedelta for human consumption. Also used by some reports.
    """
    if td is None:
        return ''
    hh, rem = divmod(td.days*24.0*SECS_IN_HR + td.seconds, SECS_IN_HR)
    mm, ss = divmod(rem, 60)
    if seconds or ss:
        return '%d:%02d:%02d' % (hh, mm, ss)
    else:
        return '%d:%02d' % (hh, mm)


def ticket_url(request, project, ticket_id):
    tracs = list(project.tracs)
    if not tracs:
        return None
    trac_url = tracs[0].application_uri(request)
    return '%s/ticket/%s' % (trac_url, ticket_id)


def timeentry_url(request, time_entry):
    return '/admin/TimeEntry/%s' % time_entry.id


def chunks(seq, num):
    avg = len(seq) / float(num)
    out = []
    last = 0.0
    while last < len(seq):
        out.append(seq[int(last):int(last + avg)])
        last += avg
    return out


def total_seconds(dt):
    """
    datetime.total_seconds is new in python 2.7
    """
    return dt.days*24*60*60 + dt.seconds


def time_chunks(dt, num):
    """
    Splits a datetime without splitting minutes
    """
    chunk_minutes = total_seconds(dt) // 60 // num
    chunk_dt = datetime.timedelta(seconds=chunk_minutes * 60)
    for n in range(num-1):
        yield chunk_dt
        dt -= chunk_dt
    yield dt



def time_parse(text, maximum=None, minimum=None):
    """
    Returns a timedelta object, given a string like HH:MM, HH or :MM
    """
    if not text:
        return None

    m = re.match('^([0-1]?[0-9]|2[0-4]):([0-5][0-9])$', text)
    if not m:
        raise ValueError(u'Cannot parse time (must be HH:MM)')

    hh, mm = m.groups()
    hh = int(hh)
    mm = int(mm)

    ret = datetime.timedelta(seconds=hh*60*60+mm*60)

    if maximum and ret > maximum:
        raise ValueError(u'Time value too big (must be <= %s)' % timedelta_as_human_str(maximum))

    if minimum and ret < minimum:
        raise ValueError(u'Time value too small (must be >= %s)' % timedelta_as_human_str(minimum))

    return ret



# Reverse ordering key for strings, dates or any kind of object.
# From http://stackoverflow.com/questions/11206884/how-to-write-python-sort-key-functions-for-descending-values
# In python 2.7, ReversedOrder can be written as:
#
#import functools
#@functools.total_ordering
#class ReversedOrder(object):
#    def __init__(self, value):
#        self.value = value
#    def __eq__(self, other):
#        return other.value == self.value
#    def __lt__(self, other):
#        return other.value < self.value

class ReversedOrder(object):
    def __init__(self, value):
        self.value = value
for x in ['__lt__', '__le__', '__eq__', '__ne__', '__ge__', '__gt__']:
    op = getattr(operator, x)
    setattr(ReversedOrder, x, lambda self, other, op=op: op(other.value, self.value))


def unicodelower(obj):
    """
    Case insensitive sort key.
    """
    return unicode(obj).lower()



def listwrap(obj):
    """
    Always returns a list, converting tuples and sequences, wraps single objects.
    """
    if obj is None:
        return []
    if isinstance(obj, basestring):   # strings are sequence type too
        return [obj]
    if operator.isSequenceType(obj):
        return list(obj)
    return [obj]





import os
import sys
import transaction
import base64
import beaker

from sqlalchemy import engine_from_config
from pyramid.paster import setup_logging, bootstrap

from penelope.core.models.dbsession import DBSession
from penelope.core.models import Base
from penelope.core.models.scripts import add_user, add_role
from penelope.core.models.dashboard import GlobalConfig


beaker.cache.cache_regions.update(dict(calculate_matrix={'key_length':''}))


def usage(argv):
    cmd = os.path.basename(argv[0])
    print('usage: %s <config_uri>\n'
          '(example: "%s development.ini")' % (cmd, cmd)) 
    sys.exit(1)

def main(argv=sys.argv):
    if len(argv) != 2:
        usage(argv)

    config_uri = argv[1]
    setup_logging(config_uri)
    env = bootstrap('%s#dashboard'% config_uri)
    settings = env.get('registry').settings
    engine = engine_from_config(settings, 'sa.dashboard.')
    DBSession.configure(bind=engine)
    Base.metadata.bind = engine
    Base.metadata.create_all()

    def generate_password():
        return base64.urlsafe_b64encode(os.urandom(30))

    with transaction.manager:
        session = DBSession()
        users = [
            add_user(session, u'admin@example.com', fullname='Administrator', password='admin@example.com'),
        ]

        role_admin = add_role(session, 'administrator')
        role_admin.users.append(users[0])

        add_role(session, 'external_developer')
        add_role(session, 'internal_developer')
        add_role(session, 'secretary')
        add_role(session, 'customer')
        add_role(session, 'project_manager')

        session.add(GlobalConfig(id=1))



# -*- coding: utf-8 -*-
from copy import deepcopy

from zope.interface import implements
from webhelpers.html import HTML
from fa.bootstrap import actions
from pyramid_formalchemy import resources
from formalchemy import fields

from penelope.core.forms import ModelView
from penelope.core.interfaces import IManageView

def configurate(config):
    config.formalchemy_model_view('admin',
        renderer='penelope.core.forms:templates/group_listing.pt',
        context=ModelListing,
        attr='datatable',
        request_method='GET',
        permission='listing',
        model='penelope.core.models.dashboard.Group',
        view=GroupModelView)

    config.formalchemy_model_view('admin',
        request_method='POST',
        permission='delete',
        name='delete',
        attr='delete',
        renderer='fa.bootstrap:templates/admin/edit.pt',
        model='penelope.core.models.dashboard.Group',
        view=GroupModelView)


class GroupLinkRenderer(fields.FieldRenderer):
    def render_readonly(self, **kwargs):
        group_id = self.field.value
        return HTML.A(u'edit', href=self.request.fa_url('Group', group_id, 'edit'))


class GroupModelView(ModelView):
    defaults_actions = deepcopy(actions.defaults_actions)
    defaults_actions.update(listing_buttons=actions.Actions())


    @actions.action('listing')
    def datatable(self, **kwargs):
        result = super(GroupModelView, self).datatable(**kwargs)

        fs = result['fs']
        fs['id'].is_raw_foreign_key = False       # formalchemy would not include foreign keys
        fs.configure(pk=True, readonly=True)

        fs._render_fields['id']._get_renderer = lambda: GroupLinkRenderer(fs._render_fields['id'])

        columns = ['project', 'roles', 'users', 'id']
        self.pick_columns(fs, columns)

        return dict(result,
                    columns=columns)

    def delete(self):
        """
        For Group we are always forcing to delete.
        No additional validation
        """
        return self.force_delete()


class ModelListing(resources.ModelListing):
    implements(IManageView)



import datetime
import lipsum
import random

from penelope.core.models.dashboard import User, Customer, CustomerRequest, Project, Role
from penelope.core.models.tp import TimeEntry


def add_user(session, email, fullname=None, password=None):
    model = User(email=email)
    if not password:
        password = email
    model.set_password(password)
    if fullname:
        model.fullname = fullname
    session.add(model)
    return model

def add_customer(session, name):
    model = Customer(name=name)
    session.add(model)
    return model

def add_customer_request(session, name, project_id):
    model = CustomerRequest(name=name, project_id=project_id, placement=CustomerRequest.PLACEMENT_BOARD)
    session.add(model)
    return model


def add_project(session, project_name, customer_id, author_id):
    model = Project(name=project_name,
                    customer_id=customer_id,
                    author_id=author_id)
    session.add(model)
    return model

def add_time_entry(session, description):
    model = TimeEntry(description=description)
    session.add(model)
    return model

def add_role(session, name):
    model = Role(name=name)
    session.add(model)
    return model

def populate_time_entries(session, users, projects, all_tickets):
    for author in users:
        date = datetime.date.today() - datetime.timedelta(days=365*3)
        te_num = 1

        g = lipsum.Generator()

        while date <= datetime.date.today():
            minutes = 8*60
            while minutes > 0:

                te_duration = random.choice([5, 10, 15, 20, 25, 30, 60, 90, 120, 180, 200, 240, 480])
                if te_duration > minutes:
                    te_duration = minutes

                minutes -= te_duration

                description = '%s.%d: %s' % (author.id, te_num, g.generate_sentence())

                te = add_time_entry(session, description=description)
                te.hours = datetime.timedelta(minutes=te_duration)
                te_num += 1
                te.date = date
                te.project_id = random.choice(projects).id

                if random.random()<0.96:
                    te.ticket = random.choice(all_tickets[te.project_id]).id
                else:
                    # once in a while, a time entry has no ticket
                    te.ticket = None

                te.author_id = author.id
            date += datetime.timedelta(days=1)



# -*- coding: utf-8 -*-

from copy import deepcopy
from fa.bootstrap import actions
from penelope.core.forms import ModelView
from penelope.core.forms import workflow


def configurate(config):
    config.formalchemy_model_view('admin',
                                  request_method='GET',
                                  permission='view',
                                  name='',
                                  attr='show',
                                  renderer='fa.bootstrap:templates/admin/show.pt',
                                  model='penelope.core.models.dashboard.Contract',
                                  view=ContractModelView)
    config.formalchemy_model_view('admin',
                                  request_method='GET',
                                  permission='workflow',
                                  name='goto_state',
                                  model='penelope.core.models.dashboard.Contract',
                                  view=workflow.goto_state)


contract_tabs = actions.TabsActions(actions.TabAction("show",
                                                      content="View",
                                                      permission='view',
                                                      attrs=dict(href="request.fa_url(request.model_name, request.model_id, '')")),)


class ContractModelView(ModelView):
    actions_categories = ('buttons', 'tabs')
    defaults_actions = deepcopy(actions.defaults_actions)

    def __init__(self, *args, **kwargs):
        super(ContractModelView, self).__init__(*args, **kwargs)
        self.defaults_actions.update(show_buttons=actions.Actions(actions.edit))
        cr_actions = deepcopy(contract_tabs)
        wf = workflow.change_workflow(self.context)
        if wf:
            cr_actions.append(wf)
        self.defaults_actions.update(show_tabs = cr_actions)



import os

from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
README = """penelope.core
=============

for more details visit: http://getpenelope.github.com/"""

CHANGES = open(os.path.join(here, 'CHANGES.txt')).read()

requires = [
    'Babel',
    'Beaker',
    'bleach',
    'deform',
    'deform_bootstrap',
    'distribute',
    'fa.bootstrap',
    'feedparser',
    'gdata==2.0.14',
    'gevent-socketio',
#    'gevent-psycopg2',
    'gspread',
    'js.jqgrid',
    'js.jquery_datatables==1.8.2',
    'js.jquery_timepicker_addon',
    'js.lesscss',
    'js.socketio',
    'js.xeditable',
    'jsonrpc',
    'lingua',
    'lorem-ipsum-generator',
    'lxml',
    'PasteScript',
    'penelope.trac',
    'plone.i18n',
    'profilehooks',
    'psycopg2',
    'pyramid',
    'pyramid_beaker',
    'pyramid_debugtoolbar',
    'pyramid_exclog',
    'pyramid_fanstatic',
    'pyramid_formalchemy',
    'pyramid_mailer',
    'pyramid_rpc',
    'pyramid_skins',
    'pyramid_zcml',
    'python-cjson',
    'python-dateutil==1.5',
    'python-openid>=2.0',
    'raven',
    'repoze.tm2>=1.0b1', # default_commit_veto
    'repoze.who-friendlyform',
    'repoze.who.plugins.sa',
    'repoze.who<1.9',
    'repoze.workflow',
    'sunburnt',
    'SQLAlchemy',
    'transaction',
    'Trac',
    'unittest2',
    'velruse',
    'WebError',
    'WebTest',
    'xlwt',
    'zope.interface',
    'zope.sqlalchemy',
    ]

tests_require = [
    'WebTest',
    'mock',
    'pyquery',
    'pytest',
    'pytest-cov',
    'pytest-pep8!=1.0.3',
    'pytest-xdist',
    'wsgi_intercept',
    'zope.testbrowser',
    ]


setup(name='penelope.core',
      version='2.0.3',
      description='Penelope main package',
      long_description=README + '\n\n' +  CHANGES,
      classifiers=[
        "Programming Language :: Python",
        "Framework :: Pylons",
        ],
      author='Penelope Team',
      author_email='penelopedev@redturtle.it',
      url='http://getpenelope.github.com',
      keywords='web wsgi bfg pylons pyramid',
      namespace_packages=['penelope'],
      packages=find_packages(),
      include_package_data=True,
      zip_safe=False,
      test_suite='penelope.core',
      install_requires = requires,
      tests_require=tests_require,
      entry_points = """\
      [paste.app_factory]
      main = penelope.core:main

      [paste.filter_app_factory]
      raven = penelope.core.ravenlog:sentry_filter_factory

      [fanstatic.libraries]
      por = penelope.core.fanstatic_resources:por_library
      deform_library = penelope.core.fanstatic_resources:deform_library
      deform_bootstrap_library = penelope.core.fanstatic_resources:deform_bootstrap_library

      [console_scripts]
      populate_penelope = penelope.core.scripts.populate:main
      populate_with_dummies = penelope.core.scripts.dummies:main
      import_svn = penelope.core.scripts.importsvn:main
      quality_export = penelope.core.scripts.quality_export:main
      contracts_import = penelope.core.scripts.contracts_import:main
      """,
      extras_require={
        'test': tests_require,}
      )



# -*- coding: utf-8 -*-

from copy import deepcopy

from zope.component import getMultiAdapter
from pyramid.httpexceptions import HTTPFound

from fa.bootstrap import actions

from penelope.core.interfaces import IApplicationView
from penelope.core.lib.helpers import unicodelower
from penelope.core.forms import ModelView
from penelope.core.models import DBSession
from penelope.core.models.dashboard import Role, ApplicationACL



def configurate(config):
    #custom view for application
    config.formalchemy_model_view('admin',
                                  request_method='GET',
                                  permission='view',
                                  name='',
                                  attr='show',
                                  model='penelope.core.models.dashboard.Application',
                                  renderer='penelope.core.forms:templates/application_show.pt',
                                  view=ApplicationModelView)

    config.formalchemy_model_view('admin',
        request_method='POST',
        permission='delete',
        name='delete',
        attr='delete',
        renderer='fa.bootstrap:templates/admin/edit.pt',
        model='penelope.core.models.dashboard.Application',
        view=ApplicationModelView)

    config.formalchemy_model_view('admin',
                                  request_method='GET',
                                  permission='edit',
                                  name='security',
                                  attr='security',
                                  model='penelope.core.models.dashboard.Application',
                                  renderer='penelope.core.forms:templates/application_acl.pt',
                                  view=ApplicationModelView)

    config.formalchemy_model_view('admin',
                                  request_method='GET',
                                  permission='edit',
                                  name='security_edit',
                                  attr='security_edit',
                                  model='penelope.core.models.dashboard.Application',
                                  renderer='penelope.core.forms:templates/application_acl.pt',
                                  view=ApplicationModelView)

    config.formalchemy_model_view('admin',
                                  request_method='POST',
                                  permission='edit',
                                  name='security_save',
                                  attr='security_save',
                                  model='penelope.core.models.dashboard.Application',
                                  renderer='penelope.core.forms:templates/application_acl.pt',
                                  view=ApplicationModelView)



app_tabs = actions.TabsActions(actions.TabAction("show",
                                                 content="View",
                                                 permission='view',
                                                 attrs=dict(href="request.fa_url(request.model_name, request.model_id, '')")),
                               actions.TabAction("security",
                                                 content="Security",
                                                 permission='edit',
                                                 attrs=dict(href="request.fa_url(request.model_name, request.model_id, 'security')")))


security_edit = actions.UIButton(id='security_edit',
                                 content='Edit',
                                 permission='edit',
                                 _class='btn btn-info',
                                 attrs=dict(href="request.fa_url(request.model_name, request.model_id, 'security_edit')"))

security_save = actions.UIButton(id='security_save',
                                 content='Save',
                                 permission='edit',
                                 _class='btn btn-success',
                                 attrs=dict(onclick="jQuery(this).parents('form').submit();"))

security_cancel = actions.UIButton(id='security_cancel',
                                   content='Cancel',
                                   permission='edit',
                                   _class='btn',
                                   attrs=dict(href="request.fa_url(request.model_name, request.model_id, 'security')"))


class ApplicationModelView(ModelView):
    """Customizations"""
    actions_categories = ('buttons', 'tabs')
    defaults_actions = deepcopy(actions.defaults_actions)
    defaults_actions['show_tabs'] = app_tabs

    acl_permission_names = ['view', 'edit']

    acl_disabled = [
                (('project_manager', 'view'), (True, True)),
                (('project_manager', 'edit'), (True, True))
            ]


    @actions.action()
    def show(self):
        context = self.context.get_instance()
        result = super(ApplicationModelView, self).show()
        renderer = getMultiAdapter((context, self.context, self.request), IApplicationView)
        result.update(renderer.render())
        return self.render(**result)


    def _security_result(self):
        context = self.context.get_instance()
        result = super(ApplicationModelView, self).show()
        renderer = getMultiAdapter((context, self.context, self.request), IApplicationView)
        result.update(renderer.render())
        roles = sorted(DBSession.query(Role).filter(Role.id!='administrator'), key=unicodelower)
        result['roles'] = roles
        result['permission_names'] = self.acl_permission_names

        result['acl'] = dict(self.acl_disabled)
        for acl in context.acl:
            result['acl'][(acl.role.id, acl.permission_name)] = (True, False)
        return result


    @actions.action()
    def security(self):
        result = self._security_result()
        result['actions']['buttons'] = actions.Actions(security_edit)
        result['form_editing'] = False
        return self.render(**result)


    @actions.action()
    def security_edit(self):
        result = self._security_result()
        result['actions']['buttons'] = actions.Actions(security_save, security_cancel)
        result['form_editing'] = True
        return self.render(**result)


    @actions.action()
    def security_save(self):
        context = self.context.get_instance()

        for acl in context.acl:
            DBSession.delete(acl)

        for checkbox_name in self.request.POST:
            role_id, permission_name = checkbox_name.split('.')
            acl = ApplicationACL(application_id=context.id,
                                 role_id=role_id,
                                 permission_name=permission_name)
            DBSession.add(acl)

        request = self.request
        return HTTPFound(location=request.fa_url(request.model_name, request.model_id, 'security'))

    def delete(self):
        """
        For Application we are always forcing to delete.
        No additional validation
        """
        return self.force_delete()





from zope.interface import Interface
from zope.schema import Object, TextLine
from penelope.core.interfaces import IManageView


class IPorModel(Interface):
    "marker interface for all por models"


class IWorkflowEnabled(IPorModel):
    "marker interface for all workflowenabled objects"


class IContract(IPorModel):
    "marker interface for contract model"


class ICustomerRequest(IPorModel):
    "marker interface for customer request model"


class ICustomer(IPorModel):
    "marker interface for customer model"


class ITimeEntry(IPorModel):
    "marker interface for timeentry model"


class IDublinCore(IPorModel):
    "marker interfaces for dublincore models"


class IKanbanBoard(IPorModel):
    "Marker interface for kanbanboard"


class IProject(IPorModel):
    "interface that provides get_tickets method"

    def get_tickets(self, limit):
        """ """


class IRoleable(IPorModel):
    "Marker interface for models that have role mapping"


class IProjectRelated(IPorModel):
    "Object that implement this inteface need to know their project"

    project = Object(schema=IProject,
                     title=u"Related project",
                     required=True) 

    project_related_label = TextLine(title=u'Project related label')
    project_related_id = TextLine(title=u'Project related id (path)')


class ITicketStore(Interface):
    " marker interface "

    def get_tickets_for_project(self, project, request, query=None, limit=None):
        """ """

    def get_requests_from_tickets(self, project, ticket_ids, request=None):
        """ """

    def get_tickets_for_request(self, customer_request, limit=None, request=None):
        """ """

class IUser(IPorModel, IManageView):
    "Marker interface for user objects"


class IRole(IPorModel, IManageView):
    "Marker interface for role objects"


class IApplication(IProjectRelated):
    "Marker interface for applications"


class ITrac(IApplication):
    "Only for trac applications"


class ITracReport(IApplication):
    "Only for trac reports applications"


class IGoogleDocs(IApplication):
    "Only for google docs applications"


class ISVN(IApplication):
    "Only for svn applications"


class IGenericApp(IApplication):
    "Only for generic applications"



# -*- coding: utf-8 -*-
from zope.interface import implements
from copy import deepcopy

from pyramid import httpexceptions as exc
from pyramid_formalchemy import resources
from fa.bootstrap import actions
from penelope.core.interfaces import IManageView
from penelope.core.forms import ModelView
from penelope.core.forms.renderers import ProjectRelationRenderer


def configurate(config):
    #custom view for user
    config.formalchemy_model_view('admin',
        request_method='GET',
        permission='view',
        name='',
        attr='show',
        renderer='penelope.core.forms:templates/user.pt',
        model='penelope.core.models.dashboard.User',
        view=UserModelView)

    config.formalchemy_model_view('admin',
        request_method='POST',
        permission='delete',
        name='delete',
        attr='delete',
        renderer='fa.bootstrap:templates/admin/edit.pt',
        model='penelope.core.models.dashboard.User',
        view=UserModelView)

    config.formalchemy_model_view('admin',
        renderer='penelope.core.forms:templates/user_listing.pt',
        attr='datatable',
        context=ModelListing,
        request_method='GET',
        permission='listing',
        model='penelope.core.models.dashboard.User',
        view=UserModelView)

    #custom view for tokens section
    config.formalchemy_model_view('admin',
        request_method='GET',
        permission='edit',
        name='user_tokens',
        attr='user_tokens',
        renderer='penelope.core.forms:templates/user_tokens.pt',
        model='penelope.core.models.dashboard.User',
        view=UserModelView)


user_tabs = actions.Actions(actions.TabAction("show",
    content="View",
    permission='view',
    attrs=dict(href="request.fa_url(request.model_name, request.model_id, '')")),
    actions.TabAction("user_tokens",
        content="User tokens",
        permission='edit',
        attrs=dict(href="request.fa_url(request.model_name, request.model_id, 'user_tokens')")), )


class ModelListing(resources.ModelListing):
    implements(IManageView)


class UserModelView(ModelView):
    actions_categories = ('buttons', 'tabs')

    defaults_actions = deepcopy(actions.defaults_actions)
    defaults_actions.update(show_tabs=user_tabs)
    defaults_actions.update(user_tokens_tabs=user_tabs)

    @actions.action('user_tokens')
    def user_tokens(self, *args, **kwargs):
        context = self.context.get_instance()
        return self.render(user=context)

    @actions.action('listing')
    def datatable(self, **kwargs):
        result = super(UserModelView, self).datatable(**kwargs)

        fs = result['fs']
        fs.configure(pk=False, readonly=True)

        columns = ['email', 'fullname', 'roles', 'project_manager', 'active']
        self.pick_columns(fs, columns)

        fs._render_fields['project_manager']._get_renderer = lambda: ProjectRelationRenderer(fs._render_fields['project_manager'])

        return dict(result,
                    columns=columns)

    def delete(self):
        user = self.context.get_instance()
        request = self.request
        if not user.active:
            request.add_message(u'User is already inactive. You cannot remove it.', type='danger')
            raise exc.HTTPFound(location=request.fa_url('User', user.id))
        else:
            user.active = False
            request.add_message(u'User deactivated.', type='success')
            raise exc.HTTPFound(location=request.fa_url('User', user.id))



# -*- coding: utf-8 -*-

import datetime
import re
import unittest2

from penelope.core.lib.helpers import time_chunks, time_parse



class TestTimeChunks(unittest2.TestCase):

    def test_onepiece(self):
        dt = datetime.timedelta(seconds=7000)
        tc = list(time_chunks(dt, 1))
        self.assertSequenceEqual(tc, [datetime.timedelta(seconds=7000)])

    def test_fifteen(self):
        dt = datetime.timedelta(seconds=15*60)
        tc = list(time_chunks(dt, 2))
        self.assertSequenceEqual(tc,
                                 [
                                     datetime.timedelta(seconds=7*60),
                                     datetime.timedelta(seconds=8*60)
                                ])

    def test_pi(self):
        dt = datetime.timedelta(seconds=3.1415926*60*60)
        tc = list(time_chunks(dt, 7))
        self.assertSequenceEqual(tc,
                                 [
                                     datetime.timedelta(seconds=26*60),
                                     datetime.timedelta(seconds=26*60),
                                     datetime.timedelta(seconds=26*60),
                                     datetime.timedelta(seconds=26*60),
                                     datetime.timedelta(seconds=26*60),
                                     datetime.timedelta(seconds=26*60),
                                     datetime.timedelta(0, 1949, 733360),
                                ])


HH = 60*60
MM = 60

class TestTimeParse(unittest2.TestCase):

    def test_hours(self):
        self.assertRaisesRegexp(ValueError, re.escape(u'Cannot parse time (must be HH:MM)'),
                                time_parse, '3')

    def test_minutes(self):
        self.assertRaisesRegexp(ValueError, re.escape(u'Cannot parse time (must be HH:MM)'),
                                time_parse, ':30')

    def test_gibberish(self):
        self.assertRaisesRegexp(ValueError, re.escape(u'Cannot parse time (must be HH:MM)'),
                                time_parse, 'blablabla')

    def test_hhmm(self):
        self.assertEquals(time_parse('2:27'),
                          datetime.timedelta(seconds=2*HH+27*MM) )

    def test_minimum(self):
        self.assertRaisesRegexp(ValueError, re.escape(u'Time value too small (must be >= 0:01'),
                                time_parse, '0:00', minimum=datetime.timedelta(seconds=1*60))

    def test_maximum(self):
        self.assertRaisesRegexp(ValueError, re.escape(u'Time value too big (must be <= 8:00'),
                                time_parse, '8:01', maximum=datetime.timedelta(seconds=8*HH))

    def test_empty(self):
        self.assertIsNone(time_parse(''))






if __name__ == '__main__':
    unittest2.main()




import transaction

from json import loads, dumps
from pyramid.response import Response
from pyramid.view import view_config
from socketio import socketio_manage
from socketio.namespace import BaseNamespace

from penelope.core.models.dashboard import KanbanBoard, \
        BACKLOG_PRIORITY_ORDER, BACKLOG_MODIFICATION_ORDER
from penelope.core.models.dashboard import Trac, Project, CustomerRequest
from penelope.core.models import DBSession


class BoardMixin(object):

    users_online = {}

    def __init__(self, *args, **kwargs):
        super(BoardMixin, self).__init__(*args, **kwargs)
        if 'boards' not in self.session:
            self.session['boards'] = set()

    def join(self, data):
        """Lets a user join a board on a specific Namespace."""
        self.session['boards'].add(self._get_board_name(data['board_id']))
        self.users_online.setdefault(self._get_board_name(data['board_id']), set()).add(data['email'])

    def leave(self, board, email):
        try:
            return self.users_online.get(self._get_board_name(board), set()).remove(email)
        except KeyError:
            pass

    def board_users(self, board):
        if not board:
            return set()
        return list(self.users_online.get(self._get_board_name(board), set()))

    def _get_board_name(self, board):
        return self.ns_name + '_' + board

    def emit_to_board(self, board, event, *args):
        """This is sent to all in the board (in this particular Namespace)"""
        pkt = dict(type="event",
                   name=event,
                   args=args,
                   endpoint=self.ns_name)
        board_name = self._get_board_name(board)
        for sessid, socket in self.socket.server.sockets.iteritems():
            if 'boards' not in socket.session:
                continue
            if board_name in socket.session['boards']:
                socket.send_packet(pkt)

    def emit_to_board_not_me(self, board, event, *args):
        """This is sent to all in the board (in this particular Namespace)"""
        pkt = dict(type="event",
                   name=event,
                   args=args,
                   endpoint=self.ns_name)
        board_name = self._get_board_name(board)
        for sessid, socket in self.socket.server.sockets.iteritems():
            if 'boards' not in socket.session:
                continue
            if board_name in socket.session['boards'] and self.socket != socket:
                socket.send_packet(pkt)


class KanbanNamespace(BaseNamespace, BoardMixin):

    @property
    def board(self):
        return self.session.get('board')

    @property
    def email(self):
        return self.session.get('email')

    def on_join(self, data):
        self.session['board'] = data['board_id']
        self.session['email'] = data['email']
        self.join(data)

        board = DBSession().query(KanbanBoard).get(self.board)
        try:
            boards = loads(board.json)
        except (ValueError, TypeError):
            boards = []
        self.emit_to_board(self.board, "columns", {"value": boards})
        self.emit_to_board(self.board, "emails", {"value": self.board_users(self.board)})

    def recv_disconnect(self):
        if self.board and self.email:
            self.leave(self.board, self.email)
            self.emit_to_board(self.board, "emails", {"value": self.board_users(self.board)})
        self.disconnect(silent=True)

    def on_history(self, data):
        self.emit_to_board_not_me(self.board, "history", data)

    def on_board_changed(self, data):
        # cleanup data - make sure we will not save empty tasks
        for col in data:
            to_remove = []
            for n, task in enumerate(col['tasks']):
                if not 'id' in task:
                    to_remove.append(n)
            to_remove.reverse()
            for n in to_remove:
                col['tasks'].pop(n)

        with transaction.manager:
            board = DBSession().query(KanbanBoard).get(self.board)
            board.json = dumps(data)
            self.emit_to_board_not_me(self.board, "columns", {"value": data})

    def on_get_backlog(self, data):
        board = DBSession().query(KanbanBoard).get(self.board)
        try:
            boards = loads(board.json)
        except (ValueError, TypeError):
            boards = []
        existing_tickets = [[b['id'] for b in a['tasks'] if b.get('id')] for a in boards]
        existing_tickets = [item for sublist in existing_tickets for item in sublist]
        crs = dict(DBSession().query(CustomerRequest.id, CustomerRequest.name))

        backlog_tickets = [t for t in self.find_tickets(board) \
                                   if '%s_%s' % (t.trac_name, t.id) \
                                                      not in existing_tickets]

        if board.backlog_order == BACKLOG_PRIORITY_ORDER:
            priorities = {'blocker': 0,
                          'critical': 1,
                          'major': 2,
                          'minor': 3,
                          'trivial': 4}
            backlog_tickets.sort(key=lambda t: priorities[t.priority])

        elif board.backlog_order == BACKLOG_MODIFICATION_ORDER:
            backlog_tickets.sort(key=lambda t: t.modification, reverse=True)

        backlog_tickets = backlog_tickets[:board.backlog_limit]
        tasks = []

        for n, ticket in enumerate(backlog_tickets):
            ticket_id = '%s_%s' % (ticket.trac_name, ticket.id)
            try:
                involved = set(ticket.involved.split(','))
            except AttributeError:
                involved = set()

            involved.add(ticket.reporter)
            involved = involved.difference([ticket.owner])

            tasks.append({'id': ticket_id,
                          'project': ticket.project,
                          'customer': ticket.customer,
                          'involvedCollapsed': True,
                          'url': '%s/trac/%s/ticket/%s' % (self.request.application_url,
                                                           ticket.trac_name,
                                                           ticket.id),
                          'ticket': ticket.id,
                          'owner': ticket.owner,
                          'involved': list(involved),
                          'customerrequest': crs.get(ticket.customerrequest,''),
                          'priority': ticket.priority in ['critical', 'blocker'] and 'true' or None,
                          'summary': ticket.summary})
        self.emit("backlog", {"value": tasks})

    def viewable_tracs(self, board):
        all_tracs = DBSession().query(Trac)
        viewable_tracs = []

        if board.projects:
            for project in board.projects:
                for trac in project.tracs:
                    viewable_tracs.append(trac)
        else:
            # a list of tracs user can view:
            if not self.request.has_permission('manage', None):
                user_projects = [g.project for g in self.request.authenticated_user.groups]
                for project in user_projects:
                    for trac in project.tracs:
                        viewable_tracs.append(trac)
            else:
                viewable_tracs = all_tracs

        query = """SELECT DISTINCT '%(trac)s' AS trac_name
                   FROM "trac_%(trac)s".permission
                   WHERE username IN ('internal_developer', '%(user)s')"""

        queries = []
        for trac in viewable_tracs:
            queries.append(query % {'trac': trac.trac_name,
                                    'user': self.request.authenticated_user.email})
        sql = '\nUNION '.join(queries)
        sql += ';'
        return DBSession().execute(sql).fetchall()

    def find_tickets(self, board):
        viewable_tracs = [t.trac_name for t in self.viewable_tracs(board)]
        all_tracs = DBSession.query(Trac).join(Project).filter(Project.active).filter(Trac.trac_name.in_(viewable_tracs))
        if all_tracs.count() == 0:
            return []
        where =  board.backlog_query or "owner='%s' AND status!='closed'" % self.request.authenticated_user.email
        query = """SELECT DISTINCT '%(trac)s' AS trac_name,
                                '%(project)s' AS project,
                                '%(customer)s' AS customer,
                                ticket.id AS id,
                                ticket.summary AS summary,
                                ticket.priority AS priority,
                                ticket.owner AS owner,
                                ticket.changetime as modification,
                                ticket.reporter as reporter,
                                customerrequest.value AS customerrequest,
                                string_agg(DISTINCT change.author,',') AS involved
                                FROM "trac_%(trac)s".ticket AS ticket
                                LEFT OUTER JOIN "trac_%(trac)s".ticket_custom AS customerrequest ON ticket.id=customerrequest.ticket AND customerrequest.name='customerrequest'
                                LEFT OUTER JOIN "trac_%(trac)s".ticket_custom AS probabilita ON ticket.id=probabilita.ticket AND probabilita.name='probabilita'
                                LEFT OUTER JOIN "trac_%(trac)s".ticket_change AS change ON ticket.id=change.ticket
                                WHERE %(where)s
                                GROUP BY ticket.id, customerrequest.value
                                """
        #WHERE %(where)s AND tp.username in ('internal_developer', '%(user)s')
        queries = []
        for trac in all_tracs:
            queries.append(query % {'trac': trac.trac_name,
                                    'project': trac.project.name,
                                    'customer': trac.project.customer.name,
                                    'where': where})
        sql = '\nUNION '.join(queries)
        sql += ';'
        tracs =  DBSession().execute(sql).fetchall()
        return tracs


@view_config(route_name="socketio")
def socketio(request):
    socketio_manage(request.environ, {"/kanban": KanbanNamespace}, request=request)
    return Response('')
